@@ -2,6 +2,8 @@
 pkgSrcFilename <- function(pkgRecord) {
   if (identical(pkgRecord$source, "github"))
     paste(pkgRecord$gh_sha1, ".tar.gz", sep = "")
+  else if (pkgRecord$source %in% c("bitbucket", "gitlab"))
+    paste(pkgRecord$remote_sha, ".tar.gz", sep = "")
   else
     paste(pkgRecord$name, "_", pkgRecord$version, ".tar.gz", sep = "")
 }
@@ -21,12 +23,12 @@ isFromCranlikeRepo <- function(pkgRecord, repos) {
   if (!length(source))
     return(TRUE)
 
-  # for records that do declare a source, ensure it's not 'source' or 'github'.
+  # for records that do declare a source, ensure it's not 'source', 'github', 'bitbucket', or 'gitlab'.
   # in previous releases of packrat, we attempted to match the repository name
   # with one of the existing repositories; however, this caused issues in
   # certain environments (the names declared repositories in the lockfile, and
   # the the names of the active repositories in the R session, may not match)
-  !tolower(source) %in% c("source", "github")
+  !tolower(source) %in% c("source", "github", "bitbucket", "gitlab")
 }
 
 # Given a package record and a database of packages, check to see if
@@ -34,14 +36,21 @@ isFromCranlikeRepo <- function(pkgRecord, repos) {
 versionMatchesDb <- function(pkgRecord, db) {
   versionMatch <-
     identical(pkgRecord$version, db[pkgRecord$name,"Version"])
+
+  # For GitHub, Bitbucket, and Gitlab, we also need to check that the SHA1 is identical
+  # (the source may be updated even if the version hasn't been bumped)
   if (versionMatch && identical(pkgRecord$source, "github")) {
-    # For GitHub, we also need to check that the SHA1 is identical (the source
-    # may be updated even if the version hasn't been bumped)
     pkgDescFile <- system.file('DESCRIPTION', package = pkgRecord$name)
     installedPkgRecord <-
       inferPackageRecord(as.data.frame(readDcf(pkgDescFile)))
     versionMatch <- identical(pkgRecord$gh_sha1,
                               installedPkgRecord$gh_sha1)
+  } else if (versionMatch && pkgRecord$source %in% c("gitlab", "bitbucket")) {
+    pkgDescFile <- system.file('DESCRIPTION', package = pkgRecord$name)
+    installedPkgRecord <-
+      inferPackageRecord(as.data.frame(readDcf(pkgDescFile)))
+    versionMatch <- identical(pkgRecord$remote_sha,
+                              installedPkgRecord$remote_sha)
   }
   versionMatch
 }
@@ -68,7 +77,8 @@ getSourceForPkgRecord <- function(pkgRecord,
   if (identical(pkgRecord$source, "unknown") && !quiet) {
     stop("No sources available for package ", pkgRecord$name, ". Packrat can ",
          "find sources for packages on CRAN-like repositories and packages ",
-         "installed using devtools::install_github. TODO: local repo")
+         "installed using devtools::install_github, devtools::install_gitlab",
+         "devtools::install_bitbucket. TODO: local repo")
   }
 
   # Create the directory in which to place this package's sources
@@ -85,7 +95,10 @@ getSourceForPkgRecord <- function(pkgRecord,
     message("Fetching sources for ", pkgRecord$name, " (", pkgRecord$version,
             ") ... ", appendLF = FALSE)
   }
+
   type <- pkgRecord$source
+  if (identical(type, "CustomCRANLikeRepository"))
+    type <- "CRAN"
 
   # If this is a local source path, compress the local sources rather than
   # trying to download from an external source
@@ -130,7 +143,7 @@ getSourceForPkgRecord <- function(pkgRecord,
 
     # Attempt to detect if this is the current version of a package
     # on a CRAN-like repository
-    currentVersion <- if (pkgRecord$name %in% rownames(availablePkgs))
+    currentVersion <- if (pkgRecord$name %in% availablePkgs[, "Package"])
       availablePkgs[pkgRecord$name, "Version"]
     else
       NA
@@ -241,6 +254,12 @@ getSourceForPkgRecord <- function(pkgRecord,
                             pkgRecord$remote_username,
                             pkgRecord$remote_repo,
                             pkgRecord$remote_sha)
+
+      # Ensure the protocol is prepended
+      if (!grepl("^http", archiveUrl)) {
+        protocol <- if (identical(method, "internal")) "http" else "https"
+        archiveUrl <- paste(protocol, archiveUrl, sep = "://")
+      }
     }
 
     srczip <- tempfile(fileext = '.tar.gz')
@@ -278,20 +297,22 @@ getSourceForPkgRecord <- function(pkgRecord,
       else
         scratchDir
 
-      if (!is.null(pkgRecord$gh_subdir))
+      if (length(pkgRecord$gh_subdir))
         basedir <- file.path(basedir, pkgRecord$gh_subdir)
 
       if (!file.exists(file.path(basedir, 'DESCRIPTION'))) {
         stop('No DESCRIPTION file was found in the archive for ', pkgRecord$name)
       }
 
-      ghinfo <- as.data.frame(c(list(
-        GithubRepo = pkgRecord$gh_repo,
+      ghdata <- c(
+        GithubRepo     = pkgRecord$gh_repo,
         GithubUsername = pkgRecord$gh_username,
-        GithubRef = pkgRecord$gh_ref,
-        GithubSHA1 = pkgRecord$gh_sha1)),
-        c(GithubSubdir = pkgRecord$gh_subdir)
+        GithubRef      = pkgRecord$gh_ref,
+        GithubSHA1     = pkgRecord$gh_sha1,
+        GithubSubdir   = pkgRecord$gh_subdir
       )
+
+      ghinfo <- as.data.frame(as.list(ghdata), stringsAsFactors = FALSE)
       appendToDcf(file.path(basedir, 'DESCRIPTION'), ghinfo)
 
       file.create(file.path(pkgSrcDir, pkgSrcFile))
@@ -308,6 +329,194 @@ getSourceForPkgRecord <- function(pkgRecord,
     })
 
     type <- "GitHub"
+  } else if (identical(pkgRecord$source, "bitbucket")) {
+
+    # Prefer using https if possible. Note that 'wininet'
+    # can fail if attempting to download from an 'http'
+    # URL that redirects to an 'https' URL.
+    # https://github.com/rstudio/packrat/issues/269
+    method <- tryCatch(
+      secureDownloadMethod(),
+      error = function(e) "internal"
+    )
+
+    if (is.null(pkgRecord$remote_host) || !nzchar(pkgRecord$remote_host)) {
+      protocol <- if (identical(method, "internal")) "http" else "https"
+      pkgRecord$remote_host <- paste0(protocol, "://bitbucket.org")
+    }
+
+    # API URLs get recorded when packages are downloaded with devtools /
+    # remotes, but Packrat just wants to use 'plain' URLs when downloading
+    # package sources.
+    originalRemoteHost <- pkgRecord$remote_host
+    pkgRecord$remote_host <- sub("api.bitbucket.org/2.0", "bitbucket.org", pkgRecord$remote_host, fixed = TRUE)
+
+    fmt <- "%s/%s/%s/get/%s.tar.gz"
+    archiveUrl <- sprintf(fmt,
+                          pkgRecord$remote_host,
+                          pkgRecord$remote_username,
+                          pkgRecord$remote_repo,
+                          pkgRecord$remote_sha)
+
+    srczip <- tempfile(fileext = '.tar.gz')
+    on.exit({
+      if (file.exists(srczip))
+        unlink(srczip, recursive = TRUE)
+    }, add = TRUE)
+
+    if (canUseBitbucketDownloader()) {
+      status <- bitbucketDownload(archiveUrl, srczip)
+      if (status) {
+        message("FAILED")
+        stop("Failed to download package from URL:\n- ", shQuote(archiveUrl))
+      }
+    } else {
+      success <- downloadWithRetries(archiveUrl, destfile = srczip, quiet = TRUE, mode = "wb")
+      if (!success) {
+        message("FAILED")
+        stop("Failed to download package from URL:\n- ", shQuote(archiveUrl))
+      }
+    }
+
+
+
+    local({
+      scratchDir <- tempfile()
+      on.exit({
+        if (file.exists(scratchDir))
+          unlink(scratchDir, recursive = TRUE)
+      })
+      # untar can emit noisy warnings (e.g. "skipping pax global extended
+      # headers"); hide those
+      suppressWarnings(untar(srczip, exdir = scratchDir, tar = "internal"))
+      # Find the base directory
+      basedir <- if (length(dir(scratchDir)) == 1)
+        file.path(scratchDir, dir(scratchDir))
+      else
+        scratchDir
+
+      if (length(pkgRecord$remote_subdir))
+        basedir <- file.path(basedir, pkgRecord$remote_subdir)
+
+      if (!file.exists(file.path(basedir, 'DESCRIPTION'))) {
+        stop('No DESCRIPTION file was found in the archive for ', pkgRecord$name)
+      }
+
+      remote_info <- as.data.frame(c(list(
+        RemoteType = "bitbucket",
+        RemoteHost = originalRemoteHost,
+        RemoteRepo = pkgRecord$remote_repo,
+        RemoteUsername = pkgRecord$remote_username,
+        RemoteRef = pkgRecord$remote_ref,
+        RemoteSha = pkgRecord$remote_sha)),
+        c(RemoteSubdir = pkgRecord$remote_subdir)
+      )
+      appendToDcf(file.path(basedir, 'DESCRIPTION'), remote_info)
+
+      file.create(file.path(pkgSrcDir, pkgSrcFile))
+      dest <- normalizePath(file.path(pkgSrcDir, pkgSrcFile), winslash = '/')
+
+      # R's internal tar (which we use here for cross-platform consistency)
+      # emits warnings when there are > 100 characters in the path, due to the
+      # resulting incompatibility with older implementations of tar. This isn't
+      # relevant for our purposes, so suppress the warning.
+      in_dir(dirname(basedir),
+             suppressWarnings(tar(tarfile = dest, files = basename(basedir),
+                                  compression = 'gzip', tar = 'internal'))
+      )
+    })
+
+    type <- "Bitbucket"
+  } else if (identical(pkgRecord$source, "gitlab")) {
+
+    # Prefer using https if possible. Note that 'wininet'
+    # can fail if attempting to download from an 'http'
+    # URL that redirects to an 'https' URL.
+    # https://github.com/rstudio/packrat/issues/269
+    method <- tryCatch(
+      secureDownloadMethod(),
+      error = function(e) "internal"
+    )
+
+    if (is.null(pkgRecord$remote_host) || !nzchar(pkgRecord$remote_host)) {
+      protocol <- if (identical(method, "internal")) "http" else "https"
+      pkgRecord$remote_host <- paste0(protocol, "://gitlab.com")
+    }
+
+    fmt <- "%s/api/v4/projects/%s%%2F%s/repository/archive?sha=%s"
+    archiveUrl <- sprintf(fmt,
+                          pkgRecord$remote_host,
+                          pkgRecord$remote_username,
+                          pkgRecord$remote_repo,
+                          pkgRecord$remote_sha)
+
+    srczip <- tempfile(fileext = '.tar.gz')
+    on.exit({
+      if (file.exists(srczip))
+        unlink(srczip, recursive = TRUE)
+    }, add = TRUE)
+
+    if (canUseGitlabDownloader()) {
+      status <- gitlabDownload(archiveUrl, srczip)
+      if (status) {
+        message("FAILED")
+        stop("Failed to download package from URL:\n- ", shQuote(archiveUrl))
+      }
+    } else {
+      success <- downloadWithRetries(archiveUrl, destfile = srczip, quiet = TRUE, mode = "wb")
+      if (!success) {
+        message("FAILED")
+        stop("Failed to download package from URL:\n- ", shQuote(archiveUrl))
+      }
+    }
+
+    local({
+      scratchDir <- tempfile()
+      on.exit({
+        if (file.exists(scratchDir))
+          unlink(scratchDir, recursive = TRUE)
+      })
+      # untar can emit noisy warnings (e.g. "skipping pax global extended
+      # headers"); hide those
+      suppressWarnings(untar(srczip, exdir = scratchDir, tar = "internal"))
+      # Find the base directory
+      basedir <- if (length(dir(scratchDir)) == 1)
+        file.path(scratchDir, dir(scratchDir))
+      else
+        scratchDir
+
+      if (length(pkgRecord$remote_subdir))
+        basedir <- file.path(basedir, pkgRecord$remote_subdir)
+
+      if (!file.exists(file.path(basedir, 'DESCRIPTION'))) {
+        stop('No DESCRIPTION file was found in the archive for ', pkgRecord$name)
+      }
+
+      remote_info <- as.data.frame(c(list(
+        RemoteType = "gitlab",
+        RemoteHost = pkgRecord$remote_host,
+        RemoteRepo = pkgRecord$remote_repo,
+        RemoteUsername = pkgRecord$remote_username,
+        RemoteRef = pkgRecord$remote_ref,
+        RemoteSha = pkgRecord$remote_sha)),
+        c(RemoteSubdir = pkgRecord$remote_subdir)
+      )
+      appendToDcf(file.path(basedir, 'DESCRIPTION'), remote_info)
+
+      file.create(file.path(pkgSrcDir, pkgSrcFile))
+      dest <- normalizePath(file.path(pkgSrcDir, pkgSrcFile), winslash = '/')
+
+      # R's internal tar (which we use here for cross-platform consistency)
+      # emits warnings when there are > 100 characters in the path, due to the
+      # resulting incompatibility with older implementations of tar. This isn't
+      # relevant for our purposes, so suppress the warning.
+      in_dir(dirname(basedir),
+             suppressWarnings(tar(tarfile = dest, files = basename(basedir),
+                                  compression = 'gzip', tar = 'internal'))
+      )
+    })
+
+    type <- "Gitlab"
   }
   if (!quiet) {
     if (file.exists(file.path(pkgSrcDir, pkgSrcFile))) {
@@ -398,10 +607,8 @@ installedByPackrat <- function(pkgNames, lib.loc, default=NA) {
   })))
 }
 
-# Removes one or more packages from the app's private library and cached
-# sources.
+# Removes one or more packages from the app's private library.
 removePkgs <- function(project, pkgNames, lib.loc = libDir(project)) {
-  unlink(file.path(srcDir(project), pkgNames), recursive = TRUE)
   remove.packages(pkgNames, lib.loc)
   pkgNames
 }
@@ -476,10 +683,11 @@ installPkg <- function(pkgRecord,
 
   # Try downloading a binary (when appropriate).
   if (!(copiedFromCache || copiedFromUntrustedCache) &&
-        hasBinaryRepositories() &&
-        isFromCranlikeRepo(pkgRecord, repos) &&
-        pkgRecord$name %in% rownames(availablePackagesBinary(repos = repos)) &&
-        versionMatchesDb(pkgRecord, availablePackagesBinary(repos = repos)))
+      hasBinaryRepositories() &&
+      binaryRepositoriesEnabled() &&
+      isFromCranlikeRepo(pkgRecord, repos) &&
+      pkgRecord$name %in% availablePackagesBinary(repos = repos)[, "Package"] &&
+      versionMatchesDb(pkgRecord, availablePackagesBinary(repos = repos)))
   {
     tempdir <- tempdir()
     tryCatch({
@@ -511,7 +719,7 @@ installPkg <- function(pkgRecord,
   }
 
   if (is.null(pkgSrc)) {
-    # When installing from github or an older version, use the cached source
+    # When installing from github/bitbucket/gitlab or an older version, use the cached source
     # tarball or zip created in snapshotSources
     pkgSrc <- file.path(srcDir(project), pkgRecord$name,
                         pkgSrcFilename(pkgRecord))
@@ -534,6 +742,16 @@ installPkg <- function(pkgRecord,
       }
     }
 
+    # Infer package type; note that RSPM may deliver binary packages
+    # in archives with .tar.gz extension.
+    pkgType <- tryCatch(
+      archivePackageType(pkgSrc, quiet = TRUE),
+      error = identity
+    )
+
+    if (identical(pkgType, "binary"))
+      type <- "downloaded binary"
+
     local({
       # devtools does not install to any libraries other than the default, so
       # if the library we wish to install to is not the default, set as the
@@ -829,3 +1047,48 @@ overlaySourcePackages <- function(srcDir, overlayDir = NULL) {
     file.exists(target)
   })
 }
+
+archivePackageType <- function(path, quiet = FALSE, default = "source") {
+
+  info <- file.info(path, extra_cols = FALSE)
+  if (is.na(info$isdir))
+    stopf("no package at path %s", shQuote(path, type = "cmd"))
+
+  # for directories, check for Meta
+  if (info$isdir) {
+    hasmeta <- file.exists(file.path(path, "Meta"))
+    type <- if (hasmeta) "binary" else "source"
+    return(type)
+  }
+
+  # otherwise, guess based on contents of package
+  methods <- list(
+    tar = function(path) untar(tarfile = path, list = TRUE),
+    zip = function(path) unzip(zipfile = path, list = TRUE)$Name
+  )
+
+  # try zip first for files ending with '.zip'
+  # (but attempt to be robust against mis-named files)
+  if (endswith(path, ".zip"))
+    methods <- methods[c("zip", "tar")]
+
+  for (method in methods) {
+
+    files <- tryCatch(method(path), error = identity)
+    if (inherits(files, "error"))
+      next
+
+    hasmeta <- any(grepl("^[^/]+/Meta/?$", files))
+    type <- if (hasmeta) "binary" else "source"
+    return(type)
+
+  }
+
+  if (!quiet) {
+    fmt <- "failed to determine type of package '%s'; assuming source"
+    warningf(fmt, shQuote(path, type = "cmd"))
+  }
+
+  default
+
+}
